#ifndef WIN32
#define _GNU_SOURCE
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <stdbool.h>
#include <popt.h>
#include <float.h>
#include <libharkio3.h>
#include "harktool.h"

void help(void)
{
    printf("\n");
    printf("\n");
    printf("NAME:\n");
    printf("  harktoolcli-genposxml\n");
    printf("\n");
    printf("Description:\n");
    printf("   create a position tag of xml file\n");
    printf("\n");
    printf("Usage: harktoolcli-genposxml [coordinate model] [options]\n");
    printf("\n");
    printf("Usage of each coordinate model:\n");
    printf("   harktoolcli-genposxml circle   --help\n");
    printf("   harktoolcli-genposxml grid     --help\n");
    printf("   harktoolcli-genposxml cylinder --help\n");
    printf("   harktoolcli-genposxml ball     --help\n");
    printf("\n");
    printf("\n");
}

bool hasParsedOpt = false;
void helpFormodel(poptContext con, const char* coordstr, int exitCode) {
    if (hasParsedOpt)
        return;
    printf("\n");
    printf("\n");
    printf("NAME:\n");
    printf("  harktoolcli-genposxml\n");
    printf("\n");
    printf("Description:\n");
    printf("   create a position tag of xml file\n");
    printf("\n");
    printf("coordinate model is %s:\n", coordstr);
    printf("\n");
    poptPrintHelp(con, stdout, 0);
    printf("\n");
    printf("\n");
    exit(exitCode);
}

//global
typedef struct {
    float min;
    float max;
    float interval;
} arg_t;

typedef struct {
    char* modelname;
    char* posestype;
    char* output;
    int  id;
    arg_t azimuth;
    arg_t elevation;
    arg_t height;
    arg_t radius;
    arg_t x_axis;
    arg_t y_axis;
    arg_t z_axis;
    char* path;
    harkio_Positions *poses;
} genposparam_t;

// values for formatting path str (do nothing if null)
typedef struct {
    float azimuth;
    float elevation;
    float height;
    float radius;
    float x_axis;
    float y_axis;
    float z_axis;
} fmtval_t;

//Prototype
void makePositionsFromModel(genposparam_t* m, const int* channels, int numChannel);
void polarToCartesian(harkio_Position* pos);
void cylindricalToCartesian(harkio_Position* pos);
void makeFilepath(const char *input, const fmtval_t* fvals, char *output);
int validateFormatStr(const char* fmt);

typedef enum { mR = 1, MR, IR, VR, mE, ME, IE, VE, mH, MH, IH, VH, mA, MA, IA, VA, mX, MX, IX, VX, mY, MY, IY, VY, mZ, MZ, IZ, VZ } optret_t;

//static void ballArgs(poptContext con,
//        enum poptCallbackReason reason,
//        /*@null@*/ const struct poptOption * opt,
//        /*@null@*/ const char * arg,
//        /*@null@*/ const void * data) {
//}
//
//static void circleArgs(poptContext con,
//        enum poptCallbackReason reason,
//        /*@null@*/ const struct poptOption * opt,
//        /*@null@*/ const char * arg,
//        /*@null@*/ const void * data) {
//}
//
//static void cylinderArgs(poptContext con,
//        enum poptCallbackReason reason,
//        /*@null@*/ const struct poptOption * opt,
//        /*@null@*/ const char * arg,
//        /*@null@*/ const void * data) {
//}
//
//static void gridArgs(poptContext con,
//        enum poptCallbackReason reason,
//        /*@null@*/ const struct poptOption * opt,
//        /*@null@*/ const char * arg,
//        /*@null@*/ const void * data) {
//}

int main(int argc, char **argv)
{
    genposparam_t prm = {
            NULL,            // modelname
            "soundsource",   // posestype
            NULL,            // output
            0,               // id
            {NAN, NAN, NAN}, // azimuth
            {NAN, NAN, NAN}, // elevation
            {NAN, NAN, NAN}, // height
            {NAN, NAN, NAN}, // radius
            {NAN, NAN, NAN}, // x_axis
            {NAN, NAN, NAN}, // y_axis
            {NAN, NAN, NAN}, // z_axis
            NULL,            // path
            NULL             // poses
    };
    char* channelsStr = NULL;
    bool selectedChannels[MAX_CH] = {}; // initialize false
    int* channels = NULL;
    int numChannel = 0;

    struct poptOption optMinRadius = { "min-radius", '\0', POPT_ARG_FLOAT, &(prm.radius.min), mR, "Minimum of radius", NULL };
    struct poptOption optMaxRadius = { "max-radius", '\0', POPT_ARG_FLOAT, &(prm.radius.max), MR, "Maximum of radius", NULL };
    struct poptOption optIntRadius = { "interval-radius", '\0', POPT_ARG_FLOAT, &(prm.radius.interval), IR, "Interval of radius", NULL };
    struct poptOption optMinElevation = { "min-elevation", '\0', POPT_ARG_FLOAT, &(prm.elevation.min), mE, "Minimum of elevation", NULL };
    struct poptOption optMaxElevation = { "max-elevation", '\0', POPT_ARG_FLOAT, &(prm.elevation.max), ME, "Maximum of elevation", NULL };
    struct poptOption optIntElevation = { "interval-elevation", '\0', POPT_ARG_FLOAT, &(prm.elevation.interval), IE, "Interval of elevation", NULL };
    struct poptOption optValElevation = { "elevation", '\0', POPT_ARG_FLOAT, &(prm.elevation.min), VE, "Value of elevation", NULL };
    struct poptOption optMinHeight = { "min-height", '\0', POPT_ARG_FLOAT, &(prm.height.min), mH, "Minimum of height", NULL };
    struct poptOption optMaxHeight = { "max-height", '\0', POPT_ARG_FLOAT, &(prm.height.max), MH, "Maximum of height", NULL };
    struct poptOption optIntHeight = { "interval-height", '\0', POPT_ARG_FLOAT, &(prm.height.interval), IH, "Interval of height", NULL };
    struct poptOption optMinAzimuth = { "min-azimuth", '\0', POPT_ARG_FLOAT, &(prm.azimuth.min), mA, "Minimum of azimuth", NULL };
    struct poptOption optMaxAzimuth = { "max-azimuth", '\0', POPT_ARG_FLOAT, &(prm.azimuth.max), MA, "Maximum of azimuth", NULL };
    struct poptOption optIntAzimuth = { "interval-azimuth", '\0', POPT_ARG_FLOAT, &(prm.azimuth.interval), IA, "Interval of azimuth", NULL };
    struct poptOption optMinXAxis = { "min-x_axis", '\0', POPT_ARG_FLOAT, &(prm.x_axis.min), mX, "Minimum of x_axis", NULL };
    struct poptOption optMaxXAxis = { "max-x_axis", '\0', POPT_ARG_FLOAT, &(prm.x_axis.max), MX, "Maximum of x_axis", NULL };
    struct poptOption optIntXAxis = { "interval-x_axis", '\0', POPT_ARG_FLOAT, &(prm.x_axis.interval), IX, "Interval of x_axis", NULL };
    struct poptOption optMinYAxis = { "min-y_axis", '\0', POPT_ARG_FLOAT, &(prm.y_axis.min), mY, "Minimum of y_axis", NULL };
    struct poptOption optMaxYAxis = { "max-y_axis", '\0', POPT_ARG_FLOAT, &(prm.y_axis.max), MY, "Maximum of y_axis", NULL };
    struct poptOption optIntYAxis = { "interval-y_axis", '\0', POPT_ARG_FLOAT, &(prm.y_axis.interval), IY, "Interval of y_axis", NULL };
    struct poptOption optMinZAxis = { "min-z_axis", '\0', POPT_ARG_FLOAT, &(prm.z_axis.min), mZ, "Minimum of z_axis", NULL };
    struct poptOption optMaxZAxis = { "max-z_axis", '\0', POPT_ARG_FLOAT, &(prm.z_axis.max), MZ, "Maximum of z_axis", NULL };
    struct poptOption optIntZAxis = { "interval-z_axis", '\0', POPT_ARG_FLOAT, &(prm.z_axis.interval), IZ, "Interval of z_axis", NULL };

    // Options for "ball"
    struct poptOption ballOptions[] = {
      /*{ "longname", "shortname",
       *  argInfo,
       *  *arg, retval,
       *  description, argment description} */
//        { NULL, '\0',
//          POPT_ARG_NONE,
//          (void *)&ballArgs, '\0',
//          NULL, NULL },
        optMinRadius,
        optMaxRadius,
        optIntRadius,
        optMinElevation,
        optMaxElevation,
        optIntElevation,
        optMinAzimuth,
        optMaxAzimuth,
        optIntAzimuth,
        POPT_TABLEEND
    };

    // Options for "circle"
    struct poptOption circleOptions[] = {
      /*{ "longname", "shortname",
       *  argInfo,
       *  *arg, retval,
       *  description, argment description} */
//        { NULL, '\0',
//          POPT_ARG_NONE,
//          (void *)&circleArgs, '\0',
//          NULL, NULL },
        optMinRadius,
        optMaxRadius,
        optIntRadius,
        optValElevation,
        optMinAzimuth,
        optMaxAzimuth,
        optIntAzimuth,
        POPT_TABLEEND
    };

    // Options for "cylinder"
    struct poptOption cylinderOptions[] = {
      /*{ "longname", "shortname",
       *  argInfo,
       *  *arg, retval,
       *  description, argment description} */
//        { NULL, '\0',
//          POPT_ARG_NONE,
//          (void *)&cylinderArgs, '\0',
//          NULL, NULL },
        optMinRadius,
        optMaxRadius,
        optIntRadius,
        optMinHeight,
        optMaxHeight,
        optIntHeight,
        optMinAzimuth,
        optMaxAzimuth,
        optIntAzimuth,
        POPT_TABLEEND
    };

    // Options for "grid"
    struct poptOption gridOptions[] = {
      /*{ "longname", "shortname",
       *  argInfo,
       *  *arg, retval,
       *  description, argment description} */
//        { NULL, '\0',
//          POPT_ARG_NONE,
//          (void *)&gridArgs, '\0',
//          NULL, NULL },
        optMinXAxis,
        optMaxXAxis,
        optIntXAxis,
        optMinYAxis,
        optMaxYAxis,
        optIntYAxis,
        optMinZAxis,
        optMaxZAxis,
        optIntZAxis,
        POPT_TABLEEND
    };

    char* logLevel = "E";
    struct poptOption optionsTable[] = {
    /*{ "longname", "shortname",
     *  argInfo,
     *  *arg, retval,
     *  description, argment description} */
    { "type", 't',
      POPT_ARG_STRING | POPT_ARGFLAG_SHOW_DEFAULT,
      &(prm.posestype), 't',
      "Type of positions --type=tsp|impulse|noise|microphone|soundsource", NULL },
    { "output", 'o',
      POPT_ARG_STRING,
      &(prm.output), 'o',
      "Output filename", NULL },
    { "id", 'i',
      POPT_ARG_INT | POPT_ARGFLAG_SHOW_DEFAULT,
      &(prm.id), 'i',
      "Position id of start", NULL },
    { "path", 'p',
      POPT_ARG_STRING,
      &(prm.path), 'p',
      NULL, NULL },
    { "channels", 'c',
      POPT_ARG_STRING,
      &channelsStr, 'c',
      "Select microphone channel(s) to calculate. (zero-based, comma separated)", "INT,..." }, // used no-break-space
    { "logLevel", '\0',
      POPT_ARG_STRING | POPT_ARGFLAG_SHOW_DEFAULT,
      &logLevel, '-',
      "Log information level", "E|W|I|D" },
    { "help", '\0',
      0,
      NULL, '?',
      "Show this help message", NULL },
    { NULL, '\0',
      POPT_ARG_INCLUDE_TABLE,
      NULL, 0,
      NULL, NULL },
    POPT_TABLEEND };

#define PATHDESC(example) "Wav file path (ex. filename_" example ")"
#define OTHEROPTIONHELP(str) "" str " [OPTIONS]* arguments"
#define COORDOPTTITLE(str) "" str " arguments:"
    const char* otherOptHelp;
    // search option index for --path
    int pathOptIdx = 0;
    while (strcmp(optionsTable[pathOptIdx].longName, "path")) {
        pathOptIdx++;
    }
    // search option index for changing coordmodel
    int coordOptIdx = 0;
    while (optionsTable[coordOptIdx].argInfo != POPT_ARG_INCLUDE_TABLE) {
        coordOptIdx++;
    }

    if (argc <= 1) {
        harkio_Log_printf(harkio_Log_Error , "argument error!!");
        help();
        exit(EXIT_FAILURE);
    } else if (argc == 2) {
        help();
        if (strcmp(argv[1] , "--help") == 0) {
        	exit(EXIT_SUCCESS);
        } else {
        	exit(EXIT_FAILURE);
        }
    } else if (strcmp(argv[1], "ball") == 0) {
        prm.modelname = argv[1];
        optionsTable[pathOptIdx].descrip = PATHDESC("D{AZIMUTH:%03.1f}_E{ELEVATION:%03.1f}_R{RADIUS:%03.1f}.wav");
        optionsTable[coordOptIdx].arg = &ballOptions;
        otherOptHelp = OTHEROPTIONHELP("ball");
        optionsTable[coordOptIdx].descrip = COORDOPTTITLE("ball");
    } else if (strcmp(argv[1], "circle") == 0) {
        prm.modelname = argv[1];
        optionsTable[pathOptIdx].descrip = PATHDESC("D{AZIMUTH:%03.1f}_E{ELEVATION:%03.1f}_R{RADIUS:%03.1f}.wav");
        optionsTable[coordOptIdx].arg = &circleOptions;
        otherOptHelp = OTHEROPTIONHELP("circle");
        optionsTable[coordOptIdx].descrip = COORDOPTTITLE("circle");
    } else if (strcmp(argv[1], "cylinder") == 0) {
        prm.modelname = argv[1];
        optionsTable[pathOptIdx].descrip = PATHDESC("D{AZIMUTH:%03.1f}_H{HEIGHT:%03.1f}_R{RADIUS:%03.1f}.wav");
        optionsTable[coordOptIdx].arg = &cylinderOptions;
        otherOptHelp = OTHEROPTIONHELP("cylinder");
        optionsTable[coordOptIdx].descrip = COORDOPTTITLE("cylinder");
    } else if (strcmp(argv[1], "grid") == 0) {
        prm.modelname = argv[1];
        optionsTable[pathOptIdx].descrip = PATHDESC("X{X_AXIS:%03.1f}_Y{Y_AXIS:%03.1f}_Z{Z_AXIS:%03.1f}.wav");
        optionsTable[coordOptIdx].arg = &gridOptions;
        otherOptHelp = OTHEROPTIONHELP("grid");
        optionsTable[coordOptIdx].descrip = COORDOPTTITLE("grid");
    } else {
        harkio_Log_printf(harkio_Log_Error,"unknown coordmodel: %s", argv[1]);
        exit(EXIT_FAILURE);
    }

    poptContext optCon = poptGetContext(NULL, argc, (const char**)argv, optionsTable, 0);
    poptSetOtherOptionHelp(optCon, otherOptHelp);

    // option processing
    int c;
    while ((c = poptGetNextOpt(optCon)) >= 0) {
        switch (c) {
            char *tokenCh, *end;
            int setChCnt;
        case 't':
            if (strcmp(prm.posestype, "tsp") != 0
                    && strcmp(prm.posestype, "impulse") != 0
                    && strcmp(prm.posestype, "noise") != 0
                    && strcmp(prm.posestype, "microphone") != 0
                    && strcmp(prm.posestype, "soundsource") != 0) {
                harkio_Log_printf(harkio_Log_Error, "%s: %s",
                        poptBadOption(optCon, POPT_BADOPTION_NOALIAS), "invalid value");
                exit(EXIT_FAILURE);
            }
            hasParsedOpt = true;
            break;
        case 'i':
            if (prm.id < 0) {
                c = POPT_ERROR_BADNUMBER;
                goto OPTIONERR;
            }
            hasParsedOpt = true;
            break;
        case 'c':
            numChannel = 0;
            tokenCh = strtok(channelsStr, ", ");
            while (tokenCh != NULL) {
                long lCh = strtol(tokenCh, &end, 10);
                if (!(end && *end == '\0') || lCh > MAX_CH || lCh < 0) {
                    c = POPT_ERROR_BADNUMBER;
                    goto OPTIONERR;
                }
                if (!selectedChannels[lCh]) {
                    selectedChannels[lCh] = true;
                    numChannel++;
                }
                tokenCh = strtok(NULL, ", ");
            }
            channels = (int*)malloc(sizeof(int) * numChannel);
            setChCnt = 0;
            for (int i = 0; i < MAX_CH; i++) {
                if (selectedChannels[i]) {
                    channels[setChCnt] = i;
                    setChCnt++;
                }
                if (setChCnt >= numChannel)
                    break;
            }
            hasParsedOpt = true;
            break;
        case mR: // --min-radius
            if (prm.radius.min <= 0.0) {
                c = POPT_ERROR_BADNUMBER;
                goto OPTIONERR;
            }
            hasParsedOpt = true;
            break;
        case MR: // --max-radius
            if (prm.radius.max <= 0.0) {
                c = POPT_ERROR_BADNUMBER;
                goto OPTIONERR;
            }
            hasParsedOpt = true;
            break;
        case VE: // --elevation
            prm.elevation.max = prm.elevation.min;
            prm.elevation.interval = 0.0;
            hasParsedOpt = true;
            break;
        case '-':
            if (strcmp(logLevel, "W") == 0) {
                harkio_Log_setLevel(harkio_Log_Warning);
            } else if (strcmp(logLevel, "I") == 0) {
                harkio_Log_setLevel(harkio_Log_Info);
            } else if (strcmp(logLevel, "D") == 0) {
                harkio_Log_setLevel(harkio_Log_Debug);
            } else {
                harkio_Log_setLevel(harkio_Log_Error);
            }
            hasParsedOpt = true;
            break;
        case '?':
            // help
            helpFormodel(optCon, argv[1], EXIT_SUCCESS);
            /* no break */
        default:
            hasParsedOpt = true;
        }
    }
    if (c < -1) {
        OPTIONERR:
        /* an error occurred during option processing */
        harkio_Log_printf(harkio_Log_Error, "%s: %s",
                poptBadOption(optCon, POPT_BADOPTION_NOALIAS), poptStrerror(c));
        exit(EXIT_FAILURE);
    }

    bool isArgOK = true;
    // check essential argument
#define CHKESSENTIAL_(target,target2)    if (isnanf(prm.target.target2)) {\
                                            harkio_Log_printf(harkio_Log_Error, "missing --"#target2"-"#target"");\
                                            isArgOK = false;\
                                        }
#define CHKESSENTIAL(target)    CHKESSENTIAL_(target,min) CHKESSENTIAL_(target,max) CHKESSENTIAL_(target,interval)
    if (strcmp(argv[1], "ball") == 0) {
        CHKESSENTIAL(radius);
        CHKESSENTIAL(elevation);
        CHKESSENTIAL(azimuth);
    } else if (strcmp(argv[1], "circle") == 0) {
        CHKESSENTIAL(radius);
        if (isnanf(prm.elevation.min)) {
            harkio_Log_printf(harkio_Log_Error, "missing --elevation");
            isArgOK = false;
        }
        CHKESSENTIAL(azimuth);
    } else if (strcmp(argv[1], "cylinder") == 0) {
        CHKESSENTIAL(radius);
        CHKESSENTIAL(height);
        CHKESSENTIAL(azimuth);
    } else if (strcmp(argv[1], "grid") == 0) {
        CHKESSENTIAL(x_axis);
        CHKESSENTIAL(y_axis);
        CHKESSENTIAL(z_axis);
    }
    if (!isArgOK)
        exit(EXIT_FAILURE);

    // check min <= max
#define CHKMINMAX(target)   if (!isnanf(prm.target.min) && prm.target.min > prm.target.max ) {\
                                harkio_Log_printf(harkio_Log_Error, "should be --min-"#target" <= --max-"#target"");\
                                isArgOK = false;\
                            }
    CHKMINMAX(radius);
    CHKMINMAX(elevation);
    CHKMINMAX(height);
    CHKMINMAX(azimuth);
    CHKMINMAX(x_axis);
    CHKMINMAX(y_axis);
    CHKMINMAX(z_axis);

    // microphone cannot have channels and must be id=0
    if (strcmp(prm.posestype, "microphone") == 0) {
        if (numChannel > 0) {
            harkio_Log_printf(harkio_Log_Error,"microphone should not have --channels");
            isArgOK = false;
        }
        if (prm.id != 0) {
            harkio_Log_printf(harkio_Log_Error,"microphone should be --id == 0");
            isArgOK = false;
        }
    }

    if (!isArgOK)
        exit(EXIT_FAILURE);


    // make positions
    prm.poses = harkio_Positions_new(prm.posestype , 0, 0, NULL);

    if (prm.poses == NULL) {
        harkio_Log_printf(harkio_Log_Error,"can not harkio_Positions_new");
        exit(EXIT_FAILURE);
    }

    makePositionsFromModel(&prm, channels, numChannel);

    if (strcmp(prm.posestype, "microphone") == 0 && prm.poses->size > MAX_CH) {
        harkio_Log_printf(harkio_Log_Error,"microphone position too large. max %d", MAX_CH);
        harkio_Positions_delete(&prm.poses);
        exit(EXIT_FAILURE);
    }

    // make new XML object and write
    harkio_XML *xml;
    xml = harkio_XML_newFromData(prm.poses, NULL, NULL);

    // output
    FILE *fp;
    if (prm.output == NULL) {
        fp = stdout;
    } else if ((fp = fopen(prm.output , "w")) == NULL) {
        harkio_Log_FileOpenFailed(prm.output);
        exit(EXIT_FAILURE);
    }

    harkio_XML_write(xml, fp);

    harkio_XML_delete(&xml);

    fclose(fp);

    harkio_Log_printf(harkio_Log_Debug, "XML file created");

    fprintf(stderr, "SUCCESS\n");
    if (prm.output)
        fprintf(stderr, " output file : %s\n", prm.output);

    // free poptContext
    poptFreeContext(optCon);

    exit(EXIT_SUCCESS);
}//mainend

void makePositionsFromModel(genposparam_t* m, const int* channels, int numChannel)
{
    harkio_Position *pos ;

    //init
    int id = m->id;
    char  formatted_path[FILENAME_MAX];

    harkio_Log_printf(harkio_Log_Debug, "f_path length=%d", sizeof(formatted_path));

    float getIni(const arg_t* target) {
        return (isnanf(target->interval))?FLT_MAX:(target->interval>=0)?target->min:target->max;
    }
    float getMax(const arg_t* target) {
        return (isnanf(target->interval))?-FLT_MAX:(target->interval>=0)?target->max:target->min;
    }
    const float iniRad = getIni(&m->radius);
    const float maxRad = getMax(&m->radius);
    const float iniEle = getIni(&m->elevation);
    const float maxEle = getMax(&m->elevation);
    const float iniHei = getIni(&m->height);
    const float maxHei = getMax(&m->height);
    const float iniAzi = getIni(&m->azimuth);
    const float maxAzi = getMax(&m->azimuth);
    const float iniX = getIni(&m->x_axis);
    const float maxX = getMax(&m->x_axis);
    const float iniY = getIni(&m->y_axis);
    const float maxY = getMax(&m->y_axis);
    const float iniZ = getIni(&m->z_axis);
    const float maxZ = getMax(&m->z_axis);

    bool isChannelSet = (numChannel > 0)?false:true;
    // NOTE:for文の終了条件は以下の意味
    // if (interval > 0) { var <= max }
    // else { var >= max }
    // 1行で書くため (var - max) * interval <= 0 としている
    // interval == 0 の時は処理末尾の if (interval == 0) break;で1回だけ処理を行い、loopを抜ける
    if (strcmp(m->modelname, "circle") == 0 || strcmp(m->modelname, "ball") == 0) {
        // --------------------------------------------------------//
        //  circle
        //     方向角（最小）
        //     方向角（最大）
        //     方向角（間隔）
        //     高さ  （最小）
        //     高さ  （最大）
        //     高さ  （間隔）
        //     半径（最小）
        //     半径（最大）
        //     半径（間隔）
        // --------------------------------------------------------//
        //半径(radius)
        for (float radius=iniRad; ((radius - maxRad) * m->radius.interval) <= 0; radius+=m->radius.interval) {
            //高さ方向(elevation)
            for (float elevation=iniEle; ((elevation - maxEle) * m->elevation.interval) <= 0; elevation+=m->elevation.interval) {
                //方向角(azimuth)
                for (float azimuth=iniAzi; ((azimuth - maxAzi) * m->azimuth.interval) <= 0; azimuth+=m->azimuth.interval) {
                    //azimuth = m->azimuth.min + iazim * m->azimuth.interval;

                    harkio_Log_printf(harkio_Log_Debug, "[%d] azimuth[%f] elevation[%f] radius[%f]",id,azimuth,elevation,radius);

                    float coord[3] = {azimuth, elevation, radius};
                    fmtval_t fvals = {NAN, NAN, NAN, NAN, NAN, NAN, NAN};
                    fvals.radius = radius;
                    fvals.elevation = elevation;
                    fvals.azimuth = azimuth;
                    makeFilepath(m->path, &fvals, formatted_path);
                    pos = harkio_Position_new(id , Cartesian , coord , formatted_path);

                    if (pos == NULL) {
                        harkio_Log_printf(harkio_Log_Error,"can not harkio_Position_new") ;
                        exit(EXIT_FAILURE);
                    }
                    polarToCartesian(pos); //極座標を直交座標 に変換する

                    if(!isChannelSet) {
                        pos->channels = channels;
                        pos->numChannels = numChannel;
                        pos->channelsUse = 1;
                        isChannelSet = true;
                    }

                    harkio_Positions_append(m->poses , pos);

                    id++;
                    if (m->azimuth.interval == 0.0)
                        break;
                } // azimuth
                if(m->elevation.interval == 0.0)
                    break;
            } // elevation
            if(m->radius.interval == 0.0)
                break;
        } // radius
    }  else if (strcmp(m->modelname, "cylinder") == 0) {
        //--------------------------------------------------------//
        // cylinder
        //    方向角（最小）
        //    方向角（最大）
        //    方向角（間隔）
        //    高さ  （最小）
        //    高さ  （最大）
        //    高さ  （間隔）
        //    半径  （最小）
        //    半径  （最大）
        //    半径  （間隔）
        //--------------------------------------------------------//
        //半径(radius)
        for (float radius=iniRad; ((radius - maxRad) * m->radius.interval) <= 0; radius+=m->radius.interval) {
            //高さ方向(height)
            for (float height=iniHei; ((height - maxHei) * m->height.interval) <= 0; height+=m->height.interval) {
                //横方向(azimuth)
                for (float azimuth=iniAzi; ((azimuth - maxAzi) * m->azimuth.interval) <= 0; azimuth+=m->azimuth.interval) {

                    harkio_Log_printf(harkio_Log_Debug, "[%d] azimuth[%f] height[%f] radius[%f]",id,azimuth,height,radius);

                    float coord[3] = {azimuth, radius, height};
                    fmtval_t fvals = {NAN, NAN, NAN, NAN, NAN, NAN, NAN};
                    fvals.radius = radius;
                    fvals.height = height;
                    fvals.azimuth = azimuth;
                    makeFilepath(m->path, &fvals, formatted_path);
                    pos = harkio_Position_new(id , Cartesian , coord , formatted_path);

                    if (pos == NULL) {
                        harkio_Log_printf(harkio_Log_Error,"can not harkio_Position_new") ;
                        exit(EXIT_FAILURE);
                    }

                    cylindricalToCartesian(pos);

                    if(!isChannelSet) {
                        pos->channels = channels;
                        pos->numChannels = numChannel;
                        pos->channelsUse = 1;
                        isChannelSet = true;
                    }

                    harkio_Positions_append(m->poses , pos);

                    id++;
                    if (m->azimuth.interval == 0.0)
                        break;
                } // azimuth
                if(m->height.interval == 0.0)
                    break;
            } // height
            if(m->radius.interval == 0.0)
                break;
        } // radius
    } else if (strcmp(m->modelname, "grid") == 0) {
        // --------------------------------------------------------//
        // grid
        //     x方向（最小）
        //     x方向（最大）
        //     x方向（間隔）
        //     y方向（最小）
        //     y方向（最大）
        //     y方向（間隔）
        //     z方向（最小）
        //     z方向（最大）
        //     z方向（間隔）
        // --------------------------------------------------------//
        for (float z=iniZ; ((z - maxZ) * m->z_axis.interval) <= 0; z+=m->z_axis.interval) {
            for (float y=iniY; ((y - maxY) * m->y_axis.interval) <= 0; y+=m->y_axis.interval) {
                for (float x=iniX; ((x - maxX) * m->x_axis.interval) <= 0; x+=m->x_axis.interval) {

                    harkio_Log_printf(harkio_Log_Debug, "[%d] = (%f, %f, %f)", id, x, y, z);

                    float coord[3] = {x, y, z};
                    fmtval_t fvals = {NAN, NAN, NAN, NAN, NAN, NAN, NAN};
                    fvals.x_axis = x;
                    fvals.y_axis = y;
                    fvals.z_axis = z;
                    makeFilepath(m->path, &fvals, formatted_path);
                    pos = harkio_Position_new(id , Cartesian , coord , formatted_path);

                    if (pos == NULL) {
                        harkio_Log_printf(harkio_Log_Error,"can not harkio_Position_new") ;
                        exit(EXIT_FAILURE);
                    }

                    if(!isChannelSet) {
                        pos->channels = channels;
                        pos->numChannels = numChannel;
                        pos->channelsUse = 1;
                        isChannelSet = true;
                    }

                    harkio_Positions_append(m->poses , pos);

                    id++;
                    if (m->x_axis.interval == 0.0)
                        break;
                }//x
                if (m->y_axis.interval == 0.0)
                    break;
            }//y
            if (m->z_axis.interval == 0.0)
                break;
        }//z
    } else {
        harkio_Log_printf(harkio_Log_Error,"Unknown coordinate model");
        exit(EXIT_FAILURE);
    }
}

void polarToCartesian(harkio_Position* pos)
{
    float coord[3];
    if (pos != NULL) {
        //極座標を直交座標 に変換する
        harkio_Position_Convert(pos , coord , Cartesian);

        //positionにセット
        pos->coord[0] = coord[0];
        pos->coord[1] = coord[1];
        pos->coord[2] = coord[2];
    }
}

void cylindricalToCartesian(harkio_Position* pos)
{
    if (pos != NULL) {
        float azimuth = pos->coord[0];
        float radius = pos->coord[1];
        float height = pos->coord[2];
        float azimuth_radian = harkio_Position_convToRadian(azimuth);

        pos->coord[0] = radius * cosf(azimuth_radian);
        pos->coord[1] = radius * sinf(azimuth_radian);
        pos->coord[2] = height;
    }
}

void makeFilepath(const char*input, const fmtval_t* fvals, char* output)
{
    strcpy(output, "");
    if (input == NULL || strlen(input) == 0) {
        return;
    }

    harkio_Log_printf(harkio_Log_Debug, "input  length=%d", sizeof(input));
    harkio_Log_printf(harkio_Log_Debug, "output length=%d", sizeof(output));

    char workstr[FILENAME_MAX];
    strcpy(workstr, input);
    float fvalue;

    char *s = workstr; // s: starting point

    //                                                  // "FILE_{AZIMUTH:%03f}_{RADIUS:%03f}.wav$" (example)
    while (s != NULL && *s != '\0') {                   // "s    |       |    |                   " ("$" means '\0')
        //                                              // "|    |       |    |                   "
        char* p = strchr(s, '{');                       // "s    p       |    |                  $"
        if (p == NULL) {                                // "|    |       |    |                   "
            strncat(output, s, FILENAME_MAX-strlen(output)-1);  // (if no '{' found, copy remaining str all then finish)
            break;                                      // "|    |       |    |                   "
        }                                               // "|    |       |    |                   "
        char* q = strchr(p, ':');                       // "s    p       q    |                  $"
        char* r = strchr(p, '}');                       // "s    p       q    r                  $"

        if (q == NULL || r == NULL || r < q) {
            harkio_Log_printf(harkio_Log_Error, "path string error: invalid after '{'");
            harkio_Log_printf(harkio_Log_Error, p);
            exit(EXIT_FAILURE);
        }
        //                                              // "s    p       q    r                   "
        *p = *q = *r = '\0';                            // "FILE_$AZIMUTH$%03f$_{RADIUS:%03f}.wav$" ("$" means '\0')
        p++;                                            // "|    |       |    |                   "
        q++;                                            // "|    |       |    |                   "
        r++;                                            // "s    $p      $q   $r                 $"

        if (validateFormatStr(q) == EXIT_FAILURE) {
            harkio_Log_printf(harkio_Log_Error, "path string error: invalid format string");
            harkio_Log_printf(harkio_Log_Error, q);
            exit(EXIT_FAILURE);
        }

        // '{'前までの文字列を追加                        // "FILE_$AZIMUTH$%03f$_{RADIUS:%03f}.wav$"
        strncat(output, s,FILENAME_MAX-strlen(output)-1);   // "s    |       |    |                   "
        //                                              // (s=FILE_$)
        s = r;                                          // "|    |p      $q   $s                 $"
        //                                              // (p=AZIMUTH$,q=%03f$,
        //                                              // s=_{RADIUS:%03f}.wav$) (for next loop)

        char* formatted_val = NULL;
        if (strcmp(p ,"AZIMUTH") == 0 && !isnanf(fvals->azimuth)) {
            fvalue = fvals->azimuth;
            asprintf(&formatted_val , q , fvalue);
        } else if (strcmp(p  ,"RADIUS") == 0 && !isnanf(fvals->radius)) {
            fvalue = fvals->radius;
            asprintf(&formatted_val , q , fvalue);
        } else if (strcmp(p  ,"ELEVATION") == 0 && !isnanf(fvals->elevation)) {
            fvalue = fvals->elevation;
            asprintf(&formatted_val , q , fvalue);
        } else if (strcmp(p ,"HEIGHT") == 0 && !isnanf(fvals->height)) {
            fvalue = (int)(roundf(fvals->height * 1000));
            asprintf(&formatted_val , q , fvalue);
        } else if (strcmp(p  , "X_AXIS") == 0 && !isnanf(fvals->x_axis)) {
            fvalue = (int)(roundf(fvals->x_axis * 1000));
            asprintf(&formatted_val , q , fvalue);
        } else if (strcmp(p  , "Y_AXIS") == 0 && !isnanf(fvals->y_axis)) {
            fvalue = (int)(roundf(fvals->y_axis * 1000));
            asprintf(&formatted_val , q , fvalue);
        } else if (strcmp(p  , "Z_AXIS") == 0 && !isnanf(fvals->z_axis)) {
            fvalue = (int)(roundf(fvals->z_axis * 1000));
            asprintf(&formatted_val , q , fvalue);
        } else {
            asprintf(&formatted_val, "{%s:%s}", p, q);
        }
        strncat(output, formatted_val, FILENAME_MAX-strlen(output)-1);
        if (formatted_val)
            free(formatted_val);
    }
}

int validateFormatStr(const char* fmt)
{
    // 始端が[%] 終端が[feEgG] その他は数字か特定の記号であれば、okとする
    int valid_fmt = EXIT_FAILURE;
    int ii = 0;
    if (fmt[ii] == '%') {
        ii = 1;
        for (; ii<strlen(fmt) - 1; ii++) {                       // strlen(fmt) >= 1 == strlen('%')
            if (strchr("0123456789-. +#", fmt[ii]) == NULL) {    // fmt[ii] is digit or in {-. +#}      {
                break;
            }
        }

        if (fmt[ii] != '\0' && strchr("feEgG", fmt[ii]) != NULL) {
            if (fmt[ii + 1] == '\0') {
                valid_fmt = EXIT_SUCCESS;
            }
        }
    }
    return valid_fmt;
}
